<!--
**Note:** When your KEP is complete, all of these comment blocks should be removed.

To get started with this template:

- [X] **Pick a hosting SIG.**
  Make sure that the problem space is something the SIG is interested in taking
  up. KEPs should not be checked in without a sponsoring SIG.
- [X] **Create an issue in kubernetes/enhancements**
  When filing an enhancement tracking issue, please make sure to complete all
  fields in that template. One of the fields asks for a link to the KEP. You
  can leave that blank until this KEP is filed, and then go back to the
  enhancement and add the link.
- [X] **Make a copy of this template directory.**
  Copy this template into the owning SIG's directory and name it
  `NNNN-short-descriptive-title`, where `NNNN` is the issue number (with no
  leading-zero padding) assigned to your enhancement above.
- [X] **Fill out as much of the kep.yaml file as you can.**
  At minimum, you should fill in the "Title", "Authors", "Owning-sig",
  "Status", and date-related fields.
- [X] **Fill out this file as best you can.**
  At minimum, you should fill in the "Summary" and "Motivation" sections.
  These should be easy if you've preflighted the idea of the KEP with the
  appropriate SIG(s).
- [X] **Create a PR for this KEP.**
  Assign it to people in the SIG who are sponsoring this process.
- [ ] **Merge early and iterate.**
  Avoid getting hung up on specific details and instead aim to get the goals of
  the KEP clarified and merged quickly. The best way to do this is to just
  start with the high-level sections and fill out details incrementally in
  subsequent PRs.

Just because a KEP is merged does not mean it is complete or approved. Any KEP
marked as `provisional` is a working document and subject to change. You can
denote sections that are under active debate as follows:

```
<<[UNRESOLVED optional short context or usernames ]>>
Stuff that is being argued.
<<[/UNRESOLVED]>>
```

When editing KEPS, aim for tightly-scoped, single-topic PRs to keep discussions
focused. If you disagree with what is already in a document, open a new PR
with suggested changes.

One KEP corresponds to one "feature" or "enhancement" for its whole lifecycle.
You do not need a new KEP to move from beta to GA, for example. If
new details emerge that belong in the KEP, edit the KEP. Once a feature has become
"implemented", major changes should get new KEPs.

The canonical place for the latest set of instructions (and the likely source
of this file) is [here](/keps/NNNN-kep-template/README.md).

**Note:** Any PRs to move a KEP to `implementable`, or significant changes once
it is marked `implementable`, must be approved by each of the KEP approvers.
If none of those approvers are still appropriate, then changes to that list
should be approved by the remaining approvers and/or the owning SIG (or
SIG Architecture for cross-cutting KEPs).
-->
# KEP-2887: OpenAPI Enum Types

<!--
This is the title of your KEP. Keep it short, simple, and descriptive. A good
title can help communicate what the KEP is and should be considered as part of
any review.
-->

<!--
A table of contents is helpful for quickly jumping to sections of a KEP and for
highlighting any additional information provided beyond the standard KEP
template.

Ensure the TOC is wrapped with
  <code>&lt;!-- toc --&rt;&lt;!-- /toc --&rt;</code>
tags, and then generate with `hack/update-toc.sh`.
-->

<!-- toc -->
- [Release Signoff Checklist](#release-signoff-checklist)
- [Summary](#summary)
- [Motivation](#motivation)
  - [Goals](#goals)
  - [Non-Goals](#non-goals)
- [Proposal](#proposal)
  - [Enum Pattern](#enum-pattern)
  - [Enum Tag](#enum-tag)
  - [Notes/Constraints/Caveats (Optional)](#notesconstraintscaveats-optional)
  - [Risks and Mitigations](#risks-and-mitigations)
- [Design Details](#design-details)
  - [OpenAPI v2 Breakage](#openapi-v2-breakage)
  - [Enum Fields Pruning for Feature Disablement](#enum-fields-pruning-for-feature-disablement)
  - [Test Plan](#test-plan)
  - [Graduation Criteria](#graduation-criteria)
    - [Alpha](#alpha)
    - [Beta](#beta)
    - [Stable](#stable)
  - [Upgrade / Downgrade Strategy](#upgrade--downgrade-strategy)
  - [Version Skew Strategy](#version-skew-strategy)
- [Production Readiness Review Questionnaire](#production-readiness-review-questionnaire)
  - [Feature Enablement and Rollback](#feature-enablement-and-rollback)
  - [Rollout, Upgrade and Rollback Planning](#rollout-upgrade-and-rollback-planning)
  - [Monitoring Requirements](#monitoring-requirements)
  - [Dependencies](#dependencies)
  - [Scalability](#scalability)
  - [Troubleshooting](#troubleshooting)
- [Implementation History](#implementation-history)
- [Drawbacks](#drawbacks)
- [Alternatives](#alternatives)
  - [Do Not Break Existing Clients](#do-not-break-existing-clients)
<!-- /toc -->

## Release Signoff Checklist

<!--
**ACTION REQUIRED:** In order to merge code into a release, there must be an
issue in [kubernetes/enhancements] referencing this KEP and targeting a release
milestone **before the [Enhancement Freeze](https://git.k8s.io/sig-release/releases)
of the targeted release**.

For enhancements that make changes to code or processes/procedures in core
Kubernetes—i.e., [kubernetes/kubernetes], we require the following Release
Signoff checklist to be completed.

Check these off as they are completed for the Release Team to track. These
checklist items _must_ be updated for the enhancement to be released.
-->

Items marked with (R) are required *prior to targeting to a milestone / release*.

- [X] (R) Enhancement issue in release milestone, which links to KEP dir in [kubernetes/enhancements] (not the initial KEP PR)
- [X] (R) KEP approvers have approved the KEP status as `implementable`
- [X] (R) Design details are appropriately documented
- [X] (R) Test plan is in place, giving consideration to SIG Architecture and SIG Testing input (including test refactors)
  - [ ] e2e Tests for all Beta API Operations (endpoints)
  - [ ] (R) Ensure GA e2e tests for meet requirements for [Conformance Tests](https://github.com/kubernetes/community/blob/master/contributors/devel/sig-architecture/conformance-tests.md) 
  - [ ] (R) Minimum Two Week Window for GA e2e tests to prove flake free
- [X] (R) Graduation criteria is in place
  - [ ] (R) [all GA Endpoints](https://github.com/kubernetes/community/pull/1806) must be hit by [Conformance Tests](https://github.com/kubernetes/community/blob/master/contributors/devel/sig-architecture/conformance-tests.md) 
- [X] (R) Production readiness review completed
- [X] (R) Production readiness review approved
- [ ] "Implementation History" section is up-to-date for milestone
- [ ] User-facing documentation has been created in [kubernetes/website], for publication to [kubernetes.io]
- [ ] Supporting documentation—e.g., additional design documents, links to mailing list discussions/SIG meetings, relevant PRs/issues, release notes

<!--
**Note:** This checklist is iterative and should be reviewed and updated every time this enhancement is being considered for a milestone.
-->

[kubernetes.io]: https://kubernetes.io/
[kubernetes/enhancements]: https://git.k8s.io/enhancements
[kubernetes/kubernetes]: https://git.k8s.io/kubernetes
[kubernetes/website]: https://git.k8s.io/website

## Summary

<!--
This section is incredibly important for producing high-quality, user-focused
documentation such as release notes or a development roadmap. It should be
possible to collect this information before implementation begins, in order to
avoid requiring implementors to split their attention between writing release
notes and implementing the feature itself. KEP editors and SIG Docs
should help to ensure that the tone and content of the `Summary` section is
useful for a wide audience.

A good summary is probably at least a paragraph in length.

Both in this section and below, follow the guidelines of the [documentation
style guide]. In particular, wrap lines to a reasonable length, to make it
easier for reviewers to cite specific portions, and to minimize diff churn on
updates.

[documentation style guide]: https://github.com/kubernetes/community/blob/master/contributors/guide/style-guide.md
-->

This KEP defines Enum Types support for OpenAPI. Currently, types in our API have fields
that are actually enums but represented as plain string. This KEP proposes a `+enum`
marker to type aliases that represent enums. The OpenAPI generator will have
the capability to recognize the marker and auto-detect possible values for an enum.

## Motivation

<!--
This section is for explicitly listing the motivation, goals, and non-goals of
this KEP.  Describe why the change is important and the benefits to users. The
motivation section can optionally provide links to [experience reports] to
demonstrate the interest in a KEP within the wider Kubernetes community.

[experience reports]: https://github.com/golang/go/wiki/ExperienceReports
-->

Currently, types that are actually enums are represented as plain strings. For example,
here is the definition of `Protocol`, which must be one of `TCP`, `UDP`, and `SCTP`.
```go
package core
// ...

// Protocol defines network protocols supported for things like container ports.
type Protocol string

const (
	// ProtocolTCP is the TCP protocol.
	ProtocolTCP Protocol = "TCP"
	// ProtocolUDP is the UDP protocol.
	ProtocolUDP Protocol = "UDP"
	// ProtocolSCTP is the SCTP protocol.
	ProtocolSCTP Protocol = "SCTP"
)

// ContainerPort represents a network port in a single container
type ContainerPort struct {
  // ...

  // Required: Supports "TCP", "UDP" and "SCTP"
  // +optional
  Protocol Protocol
  
  // ...
}
```

In order to convey that the `Protocol` type is an enum, every field of the type 
have to comment its values like `ContainerPort` does. The same goes for validation - 
protocol being one of "TCP", "UDP", and "SCTP" is duplicated among the code base.

Tools and UIs can also benefit from the additional piece of information. For example,
Protocol can be shown as a dropdown of possible values instead of a textbox.
The user no longer needs to see an error of invalid value after submits the request.

### Goals

<!--
List the specific goals of the KEP. What is it trying to achieve? How will we
know that this has succeeded?
-->

- Add a marker that indicates a field as an enum.
- Make the Kubernetes OpenAPI generator recognize the marker.
- Deduce values of an enum type from its Go type definition.
- Detect and annotate enum types in all built-in types. 
- Unify the marker with that of kube-builder's enum mechanism.

### Non-Goals

<!--
What is out of scope for this KEP? Listing non-goals helps to focus discussion
and make progress.
-->

This KEP does not provide support for validation that is more complicated than an enum type
like regex or prefix. Instead, use other validation mechanism that is or will be supported.

This KEP treat any enum type as a closed enum. Any change to the set of possible values should be
treated as creating a new type and thus should follow the standard procedure of changing an API.

## Proposal

When generating OpenAPI schema, the de facto enum types should be reflected as enums in the resulting schema.

We define enum types with following properties:
- All enum types are closed, i.e., the type includes a finite number of possible values.
  However, for compatibility, a client may allow unknown values from responses.
- All enum types should be string. There can be no enum integers, floats, etc

Additionally, This KEP makes the following assumptions on the implementation:
- The canonical definition of a type is written in go and follows the standard
  Kubernetes convention.
- All enum types have their own type alias. i.e. , no raw string enums.

### Enum Pattern

The Enum Pattern of a type, appearing in a go source file, is defined as follows:
```go
package internal

// FooType defines a type alias, underlying type must be string
type FooType string

// in the same package, define constants with the alias
const (
    BarConst FooType = "Bar"
    BazConst FooType = "Baz"
)

// in one or more resources or nested objects, one of the fields is of 
// the defined alias type.

// ExampleResource should refer FooType
type ExampleResource struct {
    FooField FooType
}

```
Enum Type is the type that follows such pattern (`FooType` in the example above).
Enum Values are the possible values of the Enum Type (`Bar` and `Baz` are Enum Values for `FooType`).

### Enum Tag
The Enum Tag is the marker `// +enum` that must be in the closest comment blocks
of an enum type following Enum Pattern.

```go
package internal

// FooType define sa type alias, underlying type must be string
// +enum
type FooType string
```

An actual example looks like
```go
package core

// Protocol defines network protocols supported for things like container ports.
// +enum
type Protocol string
```

<!--
This is where we get down to the specifics of what the proposal actually is.
This should have enough detail that reviewers can understand exactly what
you're proposing, but should not include things like API designs or
implementation. What is the desired outcome and how do we measure success?.
The "Design Details" section below is for the real
nitty-gritty.
-->

### Notes/Constraints/Caveats (Optional)

<!--
What are the caveats to the proposal?
What are some important details that didn't come across above?
Go in to as much detail as necessary here.
This might be a good place to talk about core concepts and how they relate.
-->

Not all types following the Enum Pattern are actually enum types. Here is a counter-example.
```go
package core

// StorageMedium defines ways that storage can be allocated to a volume.
type StorageMedium string

// These are the valid value for StorageMedium
const (
  StorageMediumDefault         StorageMedium = ""           // use whatever the default is for the node
  StorageMediumMemory          StorageMedium = "Memory"     // use memory (tmpfs)
  StorageMediumHugePages       StorageMedium = "HugePages"  // use hugepages
  StorageMediumHugePagesPrefix StorageMedium = "HugePages-" // prefix for full medium notation HugePages-<size>
)
```

Here, `StorageMedium` can have infinite number of possible values, which disqualify it as an enum type.

The parser of Go compiler and, as a result, `gengo`, does not support type aliases. See https://github.com/kubernetes/gengo/issues/224 for details.
As of Kubernetes 1.23, we do not have any type aliases in kubernetes/kubernetes repo. Any attempt to add aliases would break the code generation.
Example: https://github.com/kubernetes/kubernetes/pull/106300

However, as of 1.23, the enum marker is the only marker to be added to a type declaration, and would be the first marker to be affected.
Until there is a fix to `gengo`, the enum generator has the following limitations:
- the enum marker must not be added to aliases
- an aliased enum type or value SHOULD NOT have comments. Otherwise, the comments will be squashed with these of the original with undefined ordering.

If the fix to `gengo` does not arrive in this release cycle, detection and enforcement the rules above will be added to the `verify-*.sh` scripts.

### Risks and Mitigations

<!--
What are the risks of this proposal, and how do we mitigate? Think broadly.
For example, consider both security and how this will impact the larger
Kubernetes ecosystem.

How will security be reviewed, and by whom?

How will UX be reviewed, and by whom?

Consider including folks who also work outside the SIG or subproject.
-->

The syntax proposed in this KEP is different from existing implementation of kube-builder.
Namely, kube-builder requires possible values in the marker itself, and is treated
as a validation.
```go
package example

// Alias
// +kubebuilder:validation:Enum=Lion;Wolf;Dragon
type Alias string
```
The two approaches are compatible, but eventually they will have to be unified.
To retain compatibility with existing users of kube-builder, the auto-deduction will be added
to kube-builder once the work is done with built-in types.

## Design Details

<!--
This section should contain enough information that the specifics of your
change are understandable. This may include API specs (though not always
required) or even code snippets. If there's any ambiguity about HOW your
proposal will be implemented, this is the place to discuss them.
-->

### OpenAPI v2 Breakage
This feature, when enabled, breaks existing code using swagger to generate client or server stub from the spec
for languages with built-in enum support, including Java and C#.
Take the following spec as an example.

```yaml
definitions:
  Foo:
    type: object
    properties:
      bar:
        type: string
        enum: 
         - A
         - B
         - C
```

The generated Java object looks like
```java
public class Foo {

  @JsonAdapter(BarEnum.Adapter.class)
  public enum BarEnum {
    A("A"),
    B("B"),
    C("C");
    // ...
  }
  private BarEnum bar = null;
}
```
If the `enum` field was not in the definition, the `Bar` field would be a String.
Maintainers of the clients should include the change in a new major release.

If, rather than update code generator or use of the generated client, a client would prefer enum fields to be strings as they were previously, a simple preprocessor
will be provided to prune the enum information similar to that of `api-server` (see below).

### Enum Fields Pruning for Feature Disablement

If the feature is disabled through the feature gate, `api-server` will remove all aspects about
enum types from the schema, making the result identical to that of the prior format.

After this feature graduates to GA, the pruning code will be removed from `api-server`.

### Test Plan

<!--
**Note:** *Not required until targeted at a release.*

Consider the following in developing a test plan for this enhancement:
- Will there be e2e and integration tests, in addition to unit tests?
- How will it be tested in isolation vs with other components?

No need to outline all of the test cases, just the general strategy. Anything
that would count as tricky in the implementation, and anything particularly
challenging to test, should be called out.

All code is expected to have adequate tests (eventually with coverage
expectations). Please adhere to the [Kubernetes testing guidelines][testing-guidelines]
when drafting this test plan.

[testing-guidelines]: https://git.k8s.io/community/contributors/devel/sig-testing/testing.md
-->

Testing should focus on `kubernetes/kube-openapi` repository. There will be 
integration tests against cases with valid and invalid definition of enum types.

In the `kubernetes/kubernetes` repository, as part of OpenAPI v3 test suites,
there should be integration and e2e tests that validate present of enum fields. 

### Graduation Criteria

#### Alpha

 - Feature implemented behind the feature gate
 - Initial tests, both in k/k and k/kube-openapi, are completed and enabled

#### Beta

 - All enums in Kubernetes built-in types have enum tags properly added.
 - Enum syntax of kube-builder updated to match that of this KEP.

#### Stable

 - OpenAPI-level validation of enum types in built-in types removed in favor of that provided by the schema. 

### Upgrade / Downgrade Strategy
Enable/disable the OpenAPIEnum feature gate.

### Version Skew Strategy
The API is still compatible even with enum types enabled.
For the conversion from v2 to v3 (see OpenAPI v3 KEP), enum types will be carried over and thus
should not change the behavior of v3.

## Production Readiness Review Questionnaire

<!--

Production readiness reviews are intended to ensure that features merging into
Kubernetes are observable, scalable and supportable; can be safely operated in
production environments, and can be disabled or rolled back in the event they
cause increased failures in production. See more in the PRR KEP at
https://git.k8s.io/enhancements/keps/sig-architecture/1194-prod-readiness.

The production readiness review questionnaire must be completed and approved
for the KEP to move to `implementable` status and be included in the release.

In some cases, the questions below should also have answers in `kep.yaml`. This
is to enable automation to verify the presence of the review, and to reduce review
burden and latency.

The KEP must have a approver from the
[`prod-readiness-approvers`](http://git.k8s.io/enhancements/OWNERS_ALIASES)
team. Please reach out on the
[#prod-readiness](https://kubernetes.slack.com/archives/CPNHUMN74) channel if
you need any help or guidance.
-->

### Feature Enablement and Rollback

<!--
This section must be completed when targeting alpha to a release.
-->

###### How can this feature be enabled / disabled in a live cluster?

<!--
Pick one of these and delete the rest.
-->

- [X] Feature gate (also fill in values in `kep.yaml`)
  - Feature gate name: `OpenAPIEnum`
  - Components depending on the feature gate: `kube-apiserver`

###### Does enabling the feature change any default behavior?

<!--
Any change of default behavior may be surprising to users or break existing
automations, so be extremely careful here.
-->

Yes. Any clients generated from the new schema will see enum fields.
For languages with enum support, OpenAPI language-level generator (swagger) will create
Language-specific enum types instead of plain strings.

###### Can the feature be disabled once it has been enabled (i.e. can we roll back the enablement)?

<!--
Describe the consequences on existing workloads (e.g., if this is a runtime
feature, can it break the existing applications?).

NOTE: Also set `disable-supported` to `true` or `false` in `kep.yaml`.
-->

Yes. By disabling the feature gate and restart `api-server`.
###### What happens if we reenable the feature if it was previously rolled back?

This feature needs access only to information immutable at run-time. There should
be no problem re-enabling it.
###### Are there any tests for feature enablement/disablement?

<!--
The e2e framework does not currently support enabling or disabling feature
gates. However, unit tests in each component dealing with managing data, created
with and without the feature, are necessary. At the very least, think about
conversion tests if API types are being modified.
-->
N/A
### Rollout, Upgrade and Rollback Planning

<!--
This section must be completed when targeting beta to a release.
-->

###### How can a rollout or rollback fail? Can it impact already running workloads?

<!--
Try to be as paranoid as possible - e.g., what if some components will restart
mid-rollout?

Be sure to consider highly-available clusters, where, for example,
feature flags will be enabled on some API servers and not others during the
rollout. Similarly, consider large clusters and how enablement/disablement
will rollout across nodes.
-->
If the API-server has multiple replicas, with some instances not yet enabling this feature,
whether the returned OpenAPI Spec contains enum types will depend on which instance handle the request
which could result in inconsistent responses for the same request.

###### What specific metrics should inform a rollback?

<!--
What signals should users be paying attention to when the feature is young
that might indicate a serious problem?
-->
Standard API Server metrics apply to this feature. For example, when
`apiserver_request_duration_seconds` is too high for `/openapi` endpoints.

###### Were upgrade and rollback tested? Was the upgrade->downgrade->upgrade path tested?

<!--
Describe manual testing that was done and the outcomes.
Longer term, we may want to require automated upgrade/rollback tests, but we
are missing a bunch of machinery and tooling and can't do that now.
-->

Enabling, disabling, and re-enabling the feature gate yields expected outcomes.
###### Is the rollout accompanied by any deprecations and/or removals of features, APIs, fields of API types, flags, etc.?

<!--
Even if applying deprecation policies, they may still surprise some users.
-->

No.

### Monitoring Requirements

<!--
This section must be completed when targeting beta to a release.
-->

###### How can an operator determine if the feature is in use by workloads?

<!--
Ideally, this should be a metric. Operations against the Kubernetes API (e.g.,
checking if there are objects with field X set) may be a last resort. Avoid
logs or events for this purpose.
-->
N/A. Workloads that query the OpenAPI specs automatically use this feature if enabled.
###### How can someone using this feature know that it is working for their instance?

<!--
For instance, if this is a pod-related feature, it should be possible to determine if the feature is functioning properly
for each individual pod.
Pick one more of these and delete the rest.
Please describe all items visible to end users below with sufficient detail so that they can verify correct enablement
and operation of this feature.
Recall that end users cannot usually observe component logs or access metrics.
-->

- [X] Other (treat as last resort)
  - Details: the returned OpenAPI Spec contains enum types.

###### What are the reasonable SLOs (Service Level Objectives) for the enhancement?

<!--
This is your opportunity to define what "normal" quality of service looks like
for a feature.

It's impossible to provide comprehensive guidance, but at the very
high level (needs more precise definitions) those may be things like:
  - per-day percentage of API calls finishing with 5XX errors <= 1%
  - 99% percentile over day of absolute value from (job creation time minus expected
    job creation time) for cron job <= 10%
  - 99.9% of /health requests per day finish with 200 code

These goals will help you determine what you need to measure (SLIs) in the next
question.
-->
This feature does not affect the SLO of API Server or any other components.

###### What are the SLIs (Service Level Indicators) an operator can use to determine the health of the service?

<!--
Pick one more of these and delete the rest.
-->

- [X] Metrics
  - Metric name: apiserver_request_duration_seconds
  - Components exposing the metric: kube-api-server

###### Are there any missing metrics that would be useful to have to improve observability of this feature?

<!--
Describe the metrics themselves and the reasons why they weren't added (e.g., cost,
implementation difficulties, etc.).
-->

### Dependencies

<!--
This section must be completed when targeting beta to a release.
-->

###### Does this feature depend on any specific services running in the cluster?

<!--
Think about both cluster-level services (e.g. metrics-server) as well
as node-level agents (e.g. specific version of CRI). Focus on external or
optional services that are needed. For example, if this feature depends on
a cloud provider API, or upon an external software-defined storage or network
control plane.

For each of these, fill in the following—thinking about running existing user workloads
and creating new ones, as well as about cluster-level services (e.g. DNS):
  - [Dependency name]
    - Usage description:
      - Impact of its outage on the feature:
      - Impact of its degraded performance or high-error rates on the feature:
-->

- `kube-apiserver`
  - Usage description:
    - Impact of its outage on the feature: The `/openapi` endpoint is unavailable
    - Impact of its degraded performance or high-error rates on the feature: The `/openapi` endpoint is degraded.
  
### Scalability

<!--
For alpha, this section is encouraged: reviewers should consider these questions
and attempt to answer them.

For beta, this section is required: reviewers must answer these questions.

For GA, this section is required: approvers should be able to confirm the
previous answers based on experience in the field.
-->

###### Will enabling / using this feature result in any new API calls?

<!--
Describe them, providing:
  - API call type (e.g. PATCH pods)
  - estimated throughput
  - originating component(s) (e.g. Kubelet, Feature-X-controller)
Focusing mostly on:
  - components listing and/or watching resources they didn't before
  - API calls that may be triggered by changes of some Kubernetes resources
    (e.g. update of object X triggers new updates of object Y)
  - periodic API calls to reconcile state (e.g. periodic fetching state,
    heartbeats, leader election, etc.)
-->

No. This feature only adds enum fields to enum types.
###### Will enabling / using this feature result in introducing new API types?

<!--
Describe them, providing:
  - API type
  - Supported number of objects per cluster
  - Supported number of objects per namespace (for namespace-scoped objects)
-->

No.
###### Will enabling / using this feature result in any new calls to the cloud provider?

<!--
Describe them, providing:
  - Which API(s):
  - Estimated increase:
-->
No.
###### Will enabling / using this feature result in increasing size or count of the existing API objects?

<!--
Describe them, providing:
  - API type(s):
  - Estimated increase in size: (e.g., new marker of size 32B)
  - Estimated amount of new objects: (e.g., new Object X for every existing Pod)
-->
No. Enum information only appears in `/openapi/` endpoints but not standard APIs.
###### Will enabling / using this feature result in increasing time taken by any operations covered by existing SLIs/SLOs?

<!--
Look at the [existing SLIs/SLOs].

Think about adding additional work or introducing new steps in between
(e.g. need to do X to start a container), etc. Please describe the details.

[existing SLIs/SLOs]: https://git.k8s.io/community/sig-scalability/slos/slos.md#kubernetes-slisslos
-->
No.
###### Will enabling / using this feature result in non-negligible increase of resource usage (CPU, RAM, disk, IO, ...) in any components?

<!--
Things to keep in mind include: additional in-memory state, additional
non-trivial computations, excessive access to disks (including increased log
volume), significant amount of data sent and/or received over network, etc.
This through this both in small and large cases, again with respect to the
[supported limits].

[supported limits]: https://git.k8s.io/community//sig-scalability/configs-and-limits/thresholds.md
-->
No. The addition of enum fields take negligible amount of memory.
### Troubleshooting

<!--
This section must be completed when targeting beta to a release.

The Troubleshooting section currently serves the `Playbook` role. We may consider
splitting it into a dedicated `Playbook` document (potentially with some monitoring
details). For now, we leave it here.
-->

###### How does this feature react if the API server and/or etcd is unavailable?
This feature is part of API server. The feature is unavailable if API server is unavailable.

###### What are other known failure modes?

<!--
For each of them, fill in the following information by copying the below template:
  - [Failure mode brief description]
    - Detection: How can it be detected via metrics? Stated another way:
      how can an operator troubleshoot without logging into a master or worker node?
    - Mitigations: What can be done to stop the bleeding, especially for already
      running user workloads?
    - Diagnostics: What are the useful log messages and their required logging
      levels that could help debug the issue?
      Not required until feature graduated to beta.
    - Testing: Are there any tests for failure mode? If not, describe why.
-->
N/A. This feature fails if and only if the API Server becomes unavailable.

###### What steps should be taken if SLOs are not being met to determine the problem?

N/A. This feature fails if and only if the API Server becomes unavailable.

## Implementation History

<!--
Major milestones in the lifecycle of a KEP should be tracked in this section.
Major milestones might include:
- the `Summary` and `Motivation` sections being merged, signaling SIG acceptance
- the `Proposal` section being merged, signaling agreement on a proposed design
- the date implementation started
- the first Kubernetes release where an initial version of the KEP was available
- the version of Kubernetes where the KEP graduated to general availability
- when the KEP was retired or superseded
-->

- 09-09-2021 `Summary` and `Motivation` sections being merged, signaling SIG acceptance
- 09-09-2021 `Proposal` section being merged, signaling agreement on a proposed design
- 09-14-2021 Enum type generator merged as `kubernetes/kube-openapi#242`
- 11-16-2021 Enum type support for OpenAPI v2 merged as #105057

## Drawbacks

<!--
Why should this KEP _not_ be implemented?
-->
This KEP will break any users that expect enum types to be plain strings. 
This is the main reason for not implementing this feature in OpenAPI v2 (see Alternative below).
## Alternatives

<!--
What other approaches did you consider, and why did you rule them out? These do
not need to be as detailed as the proposal, but should include enough
information to express the idea and why it was not acceptable.
-->

### Do Not Break Existing Clients

This change can go along with OpenAPI v3 instead of v2, which prevent the breakage mentioned above.
However, an affected client can release a new major version to introduce the breaking change itself
instead of waiting for new major version of OpenAPI. In this way, we can "make things right" and benefit
from the improvement much earlier.